import { METADATA } from "./constants";
import { DataType } from "./DataType";
import { ObjectMetadata } from "./ObjectMetadata";
import { ObjectFieldMetadata } from "./ObjectFieldMetadata";
import { ClaireError } from "..";

const getReflectedDataType = (prototype: any, propertyKey: string): DataType => {
    let t = Reflect.getMetadata("design:type", prototype, propertyKey);
    if (!t) {
        throw new ClaireError(`Cannot get reflected data type for ${prototype.constructor.name}:${propertyKey}`);
    }

    switch (t.name) {
        case "Boolean":
            return DataType.BOOLEAN;
        case "Number":
            return DataType.NUMBER;
        case "String":
            return DataType.STRING;
        case "Array":
            return DataType.ARRAY;
        default:
            return DataType.OBJECT;
    }
};

export const initObjectMetadata = <T extends ObjectMetadata>(prototype: any): T => {
    if (prototype === Object.prototype) {
        return {} as T;
    }
    if (!prototype.hasOwnProperty(METADATA)) {
        const superPrototype = initObjectMetadata<T>(Object.getPrototypeOf(prototype));
        prototype[METADATA] = { ...superPrototype, fields: superPrototype.fields ? superPrototype.fields.slice() : [] };
    }
    return prototype[METADATA];
};

export const initFieldMetadata = <T extends ObjectFieldMetadata>(prototype: any, propertyKey: string): T => {
    let meta = initObjectMetadata<ObjectMetadata>(prototype);

    let fieldIndex = meta.fields.findIndex((f) => f.name === propertyKey);
    if (fieldIndex < 0) {
        const field: ObjectFieldMetadata = {
            name: propertyKey,
            dataType: getReflectedDataType(prototype, propertyKey),
            isRequired: true,
        };
        meta.fields.push(field);
        return field as T;
    } else {
        //-- shadow clone the field
        const field = { ...meta.fields[fieldIndex] };
        meta.fields[fieldIndex] = field;
        return field as T;
    }
};

export const getObjectMetadata = <T extends ObjectMetadata>(obj?: Function): T | undefined => obj && obj.prototype[METADATA];

export const setObjectMetadata = <T extends ObjectMetadata>(obj: Function, metadata: T): T => (obj.prototype[METADATA] = metadata);
