import { Errors, GLOBAL_INJECTION_REGISTRY } from "../common/constants";
import { ClaireError } from "../system/ClaireError";
import { AbstractModel } from "../model/AbstractModel";
import { getObjectMetadata } from "../common/metadata";
import { InitableMetadata } from "../common/InitableMetadata";
import { IStartStop } from "../system/IStartStop";
import { ModelMetadata } from "../model/ModelMetadata";

interface DAGNode {
    node: any;
    from: any[];
    to: any[];
}

export const getInjectionRegistry = (): InjectionRegistry => {
    let injectionRegistry = (globalThis as any)[GLOBAL_INJECTION_REGISTRY];

    if (!injectionRegistry) {
        injectionRegistry = new InjectionRegistry();
        (globalThis as any)[GLOBAL_INJECTION_REGISTRY] = injectionRegistry;
    }
    return injectionRegistry;
};

export interface InjectableInfo<T> {
    instance: T;
    environments?: string[];
    acceptedConstructors: any[];
}

export interface InjectionTargetInfo {
    targetPrototype?: any;
    propertyKey: string;
    //-- keep the original constructor and all the modified-by-decorator versions
    fieldConstructor: any;
}

export class InjectionRegistry {
    public readonly targets: InjectionTargetInfo[] = [];
    public readonly injectables: InjectableInfo<any>[] = [];
    public readonly dependencyDAG: DAGNode[] = [];
    public readonly models: { new (): AbstractModel }[] = [];

    private store: { [key: string]: any } = {};

    public getModelById(id: string): { new (): AbstractModel } | undefined {
        return this.models.find((m) => {
            const metadata = getObjectMetadata<ModelMetadata>(m);
            return metadata?.id === id;
        });
    }

    public getInitableInjectables(): IStartStop[] {
        const initables = this.injectables
            .filter((injectable) => {
                const metadata = getObjectMetadata<InitableMetadata>(injectable.instance.constructor);
                return metadata && metadata.initable;
            })
            .map((injectable) => injectable.instance) as IStartStop[];

        //-- flat DAG then sort initable by DAG
        const sortedDAG: DAGNode[] = [];
        const startNodes = this.dependencyDAG.filter((n) => !n.from.length);

        while (startNodes.length) {
            const nodeN = startNodes.pop()!;
            sortedDAG.push(nodeN);

            nodeN.to.forEach((m) => {
                const nodeM = this.dependencyDAG.find((n) => n.node === m)!;
                //-- remove edge from nodeN to node M
                nodeM.from = nodeM.from.filter((n) => n !== nodeN.node);
                if (!nodeM.from.length) {
                    startNodes.push(nodeM);
                }
            });
        }

        if (!sortedDAG.every((n) => !n.from.length)) {
            throw new ClaireError(Errors.INJECTION_ERROR, "Circular dependency injection detected");
        }

        // sort initables by flattened DAG
        initables.sort((i1, i2) => {
            const index1 = sortedDAG.findIndex((n) => i1 instanceof n.node);
            const index2 = sortedDAG.findIndex((n) => i2 instanceof n.node);
            return index2 - index1;
        });

        return initables;
    }

    public get<T>(key: string): T {
        return this.store[key];
    }

    public set(key: string, value: any): void {
        this.store[key] = value;
    }

    public registerModel(model: { new (): AbstractModel }) {
        //-- find duplicated model
        const modelMetadata = getObjectMetadata<ModelMetadata>(model);
        const duplicatedModelIndex = this.models.findIndex((m) => {
            const metadata = getObjectMetadata<ModelMetadata>(m);
            return metadata?.id === modelMetadata?.id;
        });
        if (duplicatedModelIndex >= 0) {
            this.models[duplicatedModelIndex] = model;
        } else {
            this.models.push(model);
        }
    }

    public registerTarget(prototype: any, propertyKey: string, injecteeConstructorFn: any): void {
        let injecteeNode = this.dependencyDAG.find((n) => n.node === injecteeConstructorFn);
        if (!injecteeNode) {
            injecteeNode = { node: injecteeConstructorFn, from: [], to: [] };
            this.dependencyDAG.push(injecteeNode);
        }
        let targetNode = this.dependencyDAG.find((n) => n.node === prototype.constructor);
        if (!targetNode) {
            targetNode = { node: prototype.constructor, from: [], to: [] };
            this.dependencyDAG.push(targetNode);
        }
        injecteeNode.from.push(prototype.constructor);
        targetNode.to.push(injecteeConstructorFn);
        this.targets.push({ targetPrototype: prototype, propertyKey, fieldConstructor: injecteeConstructorFn });
    }

    public registerInjectable(constructorFn: { new (): any }, environments?: string[]): void {
        let injectable = this.injectables.find((i) => i.instance instanceof constructorFn);
        if (!injectable) {
            injectable = { instance: new constructorFn(), environments, acceptedConstructors: [constructorFn] };
            this.injectables.push(injectable);
        }
    }

    public replaceInjectable<T, K extends T>(instance: K, instanceClasses: (Function & { prototype: T })[]) {
        const index = this.injectables.findIndex((injectable) => instanceClasses.some((instanceClass) => injectable.instance instanceof instanceClass));
        if (index < 0) {
            this.injectables.push({ instance, acceptedConstructors: instanceClasses });
        } else {
            this.injectables[index].instance = instance;
        }
    }

    public filterInjectable<T>(instanceSuperClass: Function & { prototype: T }): T[] {
        return this.injectables.filter((injectable) => injectable.instance instanceof instanceSuperClass).map((injectable) => injectable.instance);
    }

    public getInjectable<T>(instanceSuperClass: Function & { prototype: T }, env?: string): T {
        const injectables = this.injectables.filter(
            (injectable) =>
                (injectable.instance instanceof instanceSuperClass || injectable.acceptedConstructors.find((c) => c === instanceSuperClass)) &&
                (!injectable.environments || !env || injectable.environments.includes(env))
        );

        if (injectables.length > 1) {
            throw new ClaireError(Errors.INJECTION_ERROR, `Multiple injectables of: ${instanceSuperClass.name}`);
        }

        if (injectables.length === 0) {
            throw new ClaireError(Errors.INJECTION_ERROR, `No matched injectable of: ${instanceSuperClass.name}`);
        }

        return injectables[0].instance;
    }

    public performInjection(currentEnv: string | undefined, ignoreMissing: boolean = false): void {
        const targets = this.targets.filter((t) => !!t.targetPrototype);
        for (const target of targets) {
            try {
                target.targetPrototype[target.propertyKey] = this.getInjectable(target.fieldConstructor, currentEnv);
            } catch (err) {
                if (!ignoreMissing) {
                    throw new ClaireError(
                        Errors.INJECTION_ERROR,
                        `No suitable injectable for: ${target.targetPrototype.constructor.name}:${target.propertyKey}`
                    );
                }
            }
        }
    }
}
