import { DataType } from "../common/DataType";
import { initObjectMetadata, initFieldMetadata, getObjectMetadata } from "../common/metadata";
import { AbstractModel } from "../model/AbstractModel";
import { ModelMetadata } from "../model/ModelMetadata";
import { ClaireError } from "../system/ClaireError";
import { DtoFieldMetadata, RangeProps, VectorProps, MimeProps } from "./DtoFieldMetadata";
import { DtoMetadata } from "./DtoMetadata";

/**
 * A simple data transfer object.
 */
export const DTO = (config?: {
    /**
     * Strict:
     * - only fields in metadata are accepted
     * - strict type checking
     */
    strict?: boolean;

    /**
     * Make all fields of metadata isRequired = false
     */
    partial?: boolean;
}) => <T extends { new (...args: any[]): Object }>(constructor: T) => {
    const metadata = initObjectMetadata<DtoMetadata>(constructor.prototype);
    metadata.strict = config?.strict;
    if (config?.partial) {
        metadata.fields.forEach((f) => (f.isRequired = false));
    }
    return constructor;
};

export const Field = (config?: { description?: string }) => (prototype: any, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    field.description = config?.description || field.description;
};

export const Required = (required?: boolean, allowNull?: boolean) => (prototype: any, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    field.isRequired = required === undefined || required;
    field.allowNull = !field.isRequired && allowNull;
};

export const ServerValue = () => (prototype: any, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    field.serverValue = true;
};

export const Integral = () => (prototype: any, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    field.isIntegral = true;
};

export const NumberRange = (config: RangeProps) => {
    return (prototype: any, propertyKey: string) => {
        const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
        field.rangeProps = config;
    };
};

export const Enum = (config: { enum: any }) => (prototype: any, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    field.enum = Object.values(config.enum);
};

export const Vector = (config?: VectorProps) => (prototype: any, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    field.vectorProps = config;
};

export const DtoClass = (dtoClass: Function & { new (...args: any): any }) => (prototype: any, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    field.dataType = DataType.OBJECT;
    field.objectTypeConstructor = dtoClass;
};

export const Relation = (model: { new (): AbstractModel }) => (prototype: any, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    const metadata = getObjectMetadata<ModelMetadata>(model);
    field.relationId = metadata?.id;
};

export const URI = (config?: MimeProps) => (prototype: any, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    if (field.dataType !== DataType.STRING) {
        throw new ClaireError(`Cannot decorate @URI on non-text field of ${prototype.constructor.name}:${propertyKey}`);
    }
    field.mimeProps = config;
};

export const Timestamp = () => (prototype: AbstractModel, propertyKey: string) => {
    const field: any = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    field.isTimestamp = true;
    field.isBigInt = true;
    field.isIntegral = true;
    field.rangeProps = { min: 0 };
};

export const BigInt = (config?: { isBigInt?: boolean }) => (prototype: AbstractModel, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    if (field.dataType !== DataType.NUMBER) {
        throw new ClaireError(`Cannot decorate @BigInt on non-numeric field of ${prototype.constructor.name}:${propertyKey}`);
    }
    field.isBigInt = config?.isBigInt;
};

export const TextField = (config?: { length?: number }) => (prototype: AbstractModel, propertyKey: string) => {
    const field = initFieldMetadata<DtoFieldMetadata>(prototype, propertyKey);
    if (config && config.length !== undefined && (config.length <= 0 || !Number.isInteger(config.length))) {
        throw new ClaireError(`@Text decorator only accepts integer value as length, provided: ${config.length}`);
    }
    field.dataType = DataType.STRING;
    field.textLength = config?.length;
};
