import { getObjectMetadata } from "../common/metadata";
import { ClaireError } from "../system/ClaireError";
import { DataType } from "../common/DataType";
import { Errors } from "../common/constants";
import { DtoMetadata } from "./DtoMetadata";

export const validateDTO = (value: any, metadata: DtoMetadata): any => {
    if (value === undefined || value === null) {
        return value;
    }

    const newValue: any = {};

    for (const field of metadata.fields) {
        let fieldValue = value[field.name];

        if (fieldValue === undefined || (fieldValue === "" && field.dataType === DataType.STRING)) {
            if (field.isRequired && !field.serverValue) {
                throw new ClaireError(Errors.VALIDATION_ERROR, `Missing required value of field: ${field.name}`);
            } else {
                fieldValue = undefined;
            }
        } else if (fieldValue === null) {
            if (!field.allowNull) {
                throw new ClaireError(Errors.VALIDATION_ERROR, `Null value is not allowed for field: ${field.name}`);
            }
        } else {
            switch (field.dataType) {
                case DataType.NUMBER:
                    if ((metadata.strict && typeof fieldValue !== "number") || isNaN(fieldValue)) {
                        throw new ClaireError(Errors.VALIDATION_ERROR, `Numeric value is required for field: ${field.name}`);
                    }
                    fieldValue = Number(fieldValue);
                    if (field.isIntegral && !Number.isInteger(fieldValue)) {
                        throw new ClaireError(Errors.VALIDATION_ERROR, `Integer value is required for field: ${field.name}`);
                    }
                    if (field.rangeProps) {
                        if (
                            (field.rangeProps.min &&
                                ((field.rangeProps.minExclusive && field.rangeProps.min >= fieldValue) ||
                                    (!field.rangeProps.minExclusive && field.rangeProps.min > fieldValue))) ||
                            (field.rangeProps.max &&
                                ((field.rangeProps.maxExclusive && field.rangeProps.max <= fieldValue) ||
                                    (!field.rangeProps.maxExclusive && field.rangeProps.max < fieldValue)))
                        ) {
                            throw new ClaireError(Errors.VALIDATION_ERROR, `Field value of range: ${field.name}`);
                        }
                    }
                    break;
                case DataType.STRING:
                    if (metadata.strict && typeof fieldValue !== "string") {
                        throw new ClaireError(Errors.VALIDATION_ERROR, `String value is required for field: ${field.name}`);
                    }
                    fieldValue = String(fieldValue);

                    break;
                case DataType.BOOLEAN:
                    if (metadata.strict && typeof fieldValue !== "boolean") {
                        throw new ClaireError(Errors.VALIDATION_ERROR, `Boolean value is required for field: ${field.name}`);
                    }
                    if (typeof fieldValue === "string") {
                        fieldValue = fieldValue === "true";
                    } else {
                        fieldValue = Boolean(fieldValue);
                    }
                    break;
                case DataType.ARRAY:
                    if (metadata.strict && !Array.isArray(fieldValue)) {
                        throw new ClaireError(Errors.VALIDATION_ERROR, `Array value is required for field: ${field.name}`);
                    }
                    if (field.vectorProps) {
                        if (
                            (field.vectorProps.minLength && fieldValue.length < field.vectorProps.minLength) ||
                            (field.vectorProps.maxLength && fieldValue.length > field.vectorProps.maxLength)
                        ) {
                            throw new ClaireError(Errors.VALIDATION_ERROR, `Array length not in range for field: ${field.name}`);
                        }

                        if (field.vectorProps.elementDataType) {
                            switch (field.vectorProps.elementDataType) {
                                case DataType.ARRAY:
                                    let index = fieldValue.findIndex((element: any) => !Array.isArray(element));
                                    if (index >= 0) {
                                        throw new ClaireError(Errors.VALIDATION_ERROR, `Element data type mismatched at: ${field.name}[${index}]`);
                                    }
                                    break;
                                case DataType.OBJECT:
                                    //-- check and validate every element in array
                                    fieldValue.forEach((element: any, index: number) => {
                                        if (typeof element !== "object") {
                                            throw new ClaireError(Errors.VALIDATION_ERROR, `Element data type mismatched at: ${field.name}[${index}]`);
                                        }
                                        if (field.objectTypeConstructor) {
                                            const metadata = getObjectMetadata(field.objectTypeConstructor);
                                            if (metadata)
                                                try {
                                                    fieldValue[index] = validateDTO(element, metadata);
                                                } catch (err) {
                                                    throw new ClaireError(
                                                        Errors.VALIDATION_ERROR,
                                                        `Validation failed for element at: ${field.name}[${index}], nested exception: ${err.name}:${err.message}`
                                                    );
                                                }
                                        }
                                    });
                                    break;
                                default:
                                    if (fieldValue.some((element: any) => typeof element !== field.vectorProps!.elementDataType)) {
                                        throw new ClaireError(Errors.VALIDATION_ERROR, `Element data type mismatched: ${field.name}`);
                                    }
                                    break;
                            }
                        }

                        if (!field.vectorProps.allowDuplicated && fieldValue.some((element: any) => fieldValue.filter((e: any) => e === element).length > 1)) {
                            throw new ClaireError(Errors.VALIDATION_ERROR, `Duplicated values found in field: ${field.name}`);
                        }

                        if (field.vectorProps.superSet && fieldValue.some((element: any) => !field.vectorProps?.superSet?.includes(element))) {
                            throw new ClaireError(Errors.VALIDATION_ERROR, `Found not allowed value in field: ${field.name}`);
                        }
                    }
                    break;
                case DataType.OBJECT:
                    if (field.objectTypeConstructor) {
                        //-- recursively call validateDTO to validate inner field
                        const metadata = getObjectMetadata(field.objectTypeConstructor);
                        if (metadata) {
                            try {
                                fieldValue = validateDTO(fieldValue, metadata);
                            } catch (err) {
                                throw new ClaireError(
                                    Errors.VALIDATION_ERROR,
                                    `Error validating: "${field.name}", nested exception: ${err.name}:${err.message}`
                                );
                            }
                        }
                    } else if (field.mimeProps) {
                        //-- check mime
                        const file: any = fieldValue;
                        if (file && field.mimeProps.maxFileKB && file.size > field.mimeProps.maxFileKB * 1024) {
                            throw new ClaireError(
                                Errors.VALIDATION_ERROR,
                                `Max file size exceeded: ${(file.size / 1024).toFixed()} KB > ${field.mimeProps.maxFileKB} KB`
                            );
                        }
                    }
                    break;
            }
            //-- other check
            if (field.enum && !field.enum.includes(fieldValue)) {
                throw new ClaireError(Errors.VALIDATION_ERROR, `Value not belongs to enum in field: ${field.name}`);
            }
        }
        newValue[field.name] = fieldValue;
    }

    //-- if strict then extra key provided in value is not allowed
    const newValueKeys = Object.keys(newValue);
    const extraKeys = Object.keys(value).filter((key) => !newValueKeys.includes(key));

    if (extraKeys.length) {
        if (metadata.strict) {
            throw new ClaireError(Errors.VALIDATION_ERROR, `Keys not allowed: ${JSON.stringify(extraKeys)}`);
        } else {
            //-- copy extra key to new value
            extraKeys.forEach((key) => {
                newValue[key] = value[key];
            });
        }
    }

    return newValue;
};
