import { Http } from "../http/Http";
import { DataType } from "../common/DataType";
import { initFieldMetadata, initObjectMetadata, getObjectMetadata } from "../common/metadata";
import { StateMachineTransition } from "../common/StateMachineTransition";
import { getInjectionRegistry } from "../di/injection";
import { ClaireError } from "../system/ClaireError";
import { Timestamp } from "../dto/decorator";
import { AbstractModel } from "./AbstractModel";
import { ModelMetadata } from "./ModelMetadata";
import { ModelFieldMetadata } from "./ModelFieldMetadata";

export const Model = (config?: {
    /**
     * Description about this model.
     */
    description?: string;

    /**
     * Override a model by using it id so that any service that uses the old model will use this model instead.
     * To support backward compatible, this new model should extends from the old model.
     */
    override?: { new (): AbstractModel };

    /**
     * Ignore some http method to prevent auto generation of routes and/or endpoints
     */
    ignoreCrud?: Http[];
}) => <T extends { new (...args: any[]): AbstractModel }>(constructor: T) => {
    const meta = initObjectMetadata<ModelMetadata>(constructor.prototype);
    meta.id = (config?.override?.name || constructor.name).toLowerCase();
    meta.description = config?.description;
    meta.ignoreCrud = config?.ignoreCrud;
    getInjectionRegistry().registerModel(constructor);
    return constructor;
};

export const PK = (config?: { autoIncrement?: boolean }) => (prototype: AbstractModel, propertyKey: string) => {
    const field = initFieldMetadata<ModelFieldMetadata>(prototype, propertyKey);
    if (field.foreignKeyProps) {
        throw new ClaireError(`Field already declared as foreign key ${prototype.constructor.name}:${propertyKey}`);
    }

    const primaryKey = getObjectMetadata<ModelMetadata>(prototype.constructor)!.fields.find((field) => field.primaryKeyProps);
    if (primaryKey && primaryKey.name !== propertyKey) {
        throw new ClaireError(`Primary key "${primaryKey.name}" is already declared for ${prototype.constructor.name}`);
    }

    field.primaryKeyProps = {
        isAutoIncrement: config?.autoIncrement,
    };

    field.isRequired = true;
    field.serverValue = field.isIntegral = config?.autoIncrement;
};

export const FK = (config: { model: { new (): AbstractModel }; cascade?: boolean }) => (prototype: AbstractModel, propertyKey: string) => {
    const field = initFieldMetadata<ModelFieldMetadata>(prototype, propertyKey);
    if (field.primaryKeyProps) {
        throw new ClaireError(`Field already declared as primary key ${prototype.constructor.name}:${propertyKey}`);
    }

    //-- get primary key of foreign model
    const primaryKey = getObjectMetadata<ModelMetadata>(config.model)!.fields.find((field) => field.primaryKeyProps)!;
    field.isIntegral = primaryKey.isIntegral;
    field.isBigInt = primaryKey.isBigInt;
    field.textLength = primaryKey.textLength;

    field.foreignKeyProps = {
        id: config.model.name.toLowerCase(),
        cascade: config.cascade,
    };
};

export const Indexing = (config?: { group?: string }) => (prototype: AbstractModel, propertyKey: string) => {
    const field = initFieldMetadata<ModelFieldMetadata>(prototype, propertyKey);
    if (!field.indexingProps) {
        field.indexingProps = { groups: [] };
    }
    if (config?.group && field.indexingProps.groups.includes(config.group)) {
        field.indexingProps.groups.push(config.group);
    }
};

export const Unique = (name?: string) => (prototype: AbstractModel, propertyKey: string) => {
    const field = initFieldMetadata<ModelFieldMetadata>(prototype, propertyKey);
    field.uniqueProps = { groupName: name };
};

export const StateMachine = <T>(config: { transition: StateMachineTransition<T>[] }) => (prototype: AbstractModel, propertyKey: string) => {
    const field = initFieldMetadata<ModelFieldMetadata>(prototype, propertyKey);
    field.stateMachineTransition = config.transition;
};

export const Searchable = (config?: { caseSensitive?: boolean; accentSensitive?: boolean }) => (prototype: AbstractModel, propertyKey: string) => {
    const field = initFieldMetadata<ModelFieldMetadata>(prototype, propertyKey);
    if (field.dataType !== DataType.STRING) {
        throw new ClaireError(`Cannot decorate @Searchable on non-text field of ${prototype.constructor.name}:${propertyKey}`);
    }
    field.searchable = true;
    field.caseSensitive = config?.caseSensitive;
    field.accentSensitive = config?.accentSensitive;
};

export const ServerTime = () => (prototype: AbstractModel, propertyKey: string) => {
    Timestamp()(prototype, propertyKey);
    const field = initFieldMetadata<ModelFieldMetadata>(prototype, propertyKey);
    field.serverValue = true;
    field.isRequired = true;
};

export const ServerModify = <T>(config?: { fn?: (value: T) => T }) => (prototype: AbstractModel, propertyKey: string) => {
    const field = initFieldMetadata<ModelFieldMetadata>(prototype, propertyKey);
    field.serverModify = config || {};
};

export const Label = () => (prototype: AbstractModel, propertyKey: string) => {
    const field = initFieldMetadata<ModelFieldMetadata>(prototype, propertyKey);
    field.isLabel = true;
};
